#include <new>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <squirrel.h>
#include <sqthread.h>

#include <AS3.h>
#include "sqwrapper.h"

extern void sqas_init();
extern int  sqas_drive(int diff);
extern void sqas_done();

/**
 * 
 */
static AS3_Val asInit(void *data, AS3_Val p)
{
	sqas_init();
	return NULL;
}

/**
 * XV
 */
static AS3_Val asDrive(void *data, AS3_Val args)
{
	int diff = 0;
	AS3_ArrayValue(args, "IntType", &diff);
	return AS3_Number(sqas_drive(diff));
}

/**
 * I
 */
static AS3_Val asDone(void *data, AS3_Val args)
{
	sqas_done();
	return NULL;
}

/**
 * squirrel XNvg̎s
 * @param script XNvg
 * @param ... 
 * @return s
 */
static AS3_Val asExec(void *data, AS3_Val p)
{
	AS3Variant result;
	AS3_Acquire(p);
	AS3Variant args(p);
	int len = args.get("length");
	if (len > 0) {
		AS3Variant text = args.get(0);
		if (text.getType() == AS3Variant::TYPE_STRING) {
			HSQUIRRELVM gv = sqobject::getGlobalVM();
			const char *textStr = text;
			if (SQ_SUCCEEDED(sq_compilebuffer(gv, textStr, strlen(textStr), _SC("TEXT"), SQTrue))) {
				sq_pushroottable(gv); // this
				for (int i=1; i<len; i++) {
					AS3Variant arg = args.get(i);
					sq_pushvariant(gv, arg);
				}
				if (SQ_SUCCEEDED(sq_call(gv, len, SQTrue, SQTrue))) {
					result = createVariant(gv, -1); // result
					sq_pop(gv, 2); // result, func
				} else {
					sq_pop(gv, 1); // func
				}
			} else {
				//SQEXCEPTION(gv);
			}
		} else {
			//SQEXCEPTION(gv);
		}
	}
	return result.getObj();
}

/**
 * squirrel O[o\bȟĂяo
 * @param name \bh
 * @param ... 
 * @return XNvg̎s
 */
static AS3_Val asCall(void *data, AS3_Val p)
{
	AS3Variant result = NULL;
	AS3_Acquire(p);
	AS3Variant args(p);
	int len = args.get("length");
	if (len > 0) {
		AS3Variant methodName = args.get(0);
		if (methodName.getType() == AS3Variant::TYPE_STRING) {
			HSQUIRRELVM gv = sqobject::getGlobalVM();
			sq_pushroottable(gv); // root
			sq_pushstring(gv, methodName, -1);
			if (SQ_SUCCEEDED(sq_get(gv, -2))) { // [ge[u֐擾
				sq_pushroottable(gv); // this
				for (int i=1;i<len;i++) {	// 
					AS3Variant arg = args.get(i);
					sq_pushvariant(gv, arg);
				}
				if (SQ_SUCCEEDED(sq_call(gv, len, SQTrue, SQTrue))) {
					result = createVariant(gv, -1); // thread
					sq_pop(gv, 3); // result, func, root
				} else {
					sq_pop(gv, 2); // func, root
					//SQEXCEPTION(gv);
				}
			} else {
				sq_pop(gv, 1); // root
				//SQEXCEPTION(gv);
			}
		}
	}
	return result.getObj();
}

/**
 * Squirrel XNvg̃XbhsB
 * @param text XNvgi[ꂽ
 * @param ... 
 */
static AS3_Val asFork(void *data, AS3_Val p)
{
	AS3_Acquire(p);
	AS3Variant args(p);
	int len = args.get("length");
	if (len > 0) {
		AS3Variant text = args.get(0);
		if (text.getType() == AS3Variant::TYPE_STRING) {
			HSQUIRRELVM gv = sqobject::getGlobalVM();
			sq_pushroottable(gv); // root
			sq_pushstring(gv, SQTHREADNAME, -1);
			if (SQ_SUCCEEDED(sq_get(gv,-2))) { // class
				sq_pushroottable(gv); // :self(root)
				sq_pushnull(gv); // delegate
				const char *textStr = text;
				if (textStr && SQ_SUCCEEDED(sq_compilebuffer(gv, textStr, strlen(textStr), _SC("TEXT"), SQTrue))) {
					for (int i=1; i<len; i++) {
						AS3Variant arg = args.get(i);
						sq_pushvariant(gv, arg);
					}
					if (SQ_SUCCEEDED(sq_call(gv, len+2, SQTrue, SQTrue))) { // RXgN^Ăяo
						sq_pop(gv, 1); // thread
					}
				} else {
					sq_pop(gv, 2); // delegate, self
				}
				sq_pop(gv, 1); // class
			}
			sq_pop(gv,1); // root
		}
	}
	return NULL;
}

/**
 * Squirrel XNvg̃t@C̃XbhsB
 * @param filename XNvgi[ꂽt@C
 * @param ... 
 */
static AS3_Val asForkStorage(void *data, AS3_Val p)
{
	AS3_Acquire(p);
	AS3Variant args(p);
	int len = args.get("length");
	if (len > 0) {
		AS3Variant scriptName = args.get(0);
		if (scriptName.getType() == AS3Variant::TYPE_STRING) {
			HSQUIRRELVM gv = sqobject::getGlobalVM();
			sq_pushroottable(gv); // root
			sq_pushstring(gv, SQTHREADNAME, -1);
			if (SQ_SUCCEEDED(sq_get(gv,-2))) { // class
				sq_pushroottable(gv); // :self(root)
				sq_pushnull(gv);      // :delegate
				sq_pushstring(gv, scriptName, -1); // :func
				for (int i=1; i<len; i++) {
					AS3Variant arg = args.get(i);
					sq_pushvariant(gv, arg);
				}
				if (SQ_SUCCEEDED(sq_call(gv, len+2, SQTrue, SQTrue))) { // RXgN^Ăяo
					//AS3Variant ret = createVariant(gv, -1); // thread
					sq_pop(gv, 1); // thread
				} else {
					sq_pop(gv, 2); // delegate, self
				}
				sq_pop(gv, 1); // class
			}
			sq_pop(gv,1); // root
		}
	}
	return NULL;
}

/**
 * Squirrel^Xbhp̃gKĂяo
 * @param name gK
 */
static AS3_Val asTrigger(void *data, AS3_Val p)
{
	AS3_Acquire(p);
	AS3Variant args(p);
	AS3Variant name = args.get(0);
	sqobject::Thread::trigger(name);
	return NULL;
}

/**
 * swcpɃC^[tF[Xo^邽߂main
 */
int
main(int argc, char **argv)
{ 
	AS3_Val init        = AS3_Function(NULL, asInit);
	AS3_Val drive       = AS3_Function(NULL, asDrive);
	AS3_Val done        = AS3_Function(NULL, asDone);
	AS3_Val exec        = AS3_Function(NULL, asExec);
	AS3_Val call        = AS3_Function(NULL, asCall);
	AS3_Val fork        = AS3_Function(NULL, asFork);
	AS3_Val forkStorage = AS3_Function(NULL, asForkStorage);
	AS3_Val trigger     = AS3_Function(NULL, asTrigger);
	AS3_Val result = AS3_Object("init:AS3ValType,"
								"drive:AS3ValType,"
								"done:AS3ValType,"
								"exec:AS3ValType,"
								"call:AS3ValType,"
								"fork:AS3ValType,"
								"forkStorage:AS3ValType,"
								"trigger:AS3ValType",
								init, drive, done, exec, call, fork, forkStorage, trigger);
	AS3_Release(init);
	AS3_Release(drive);
	AS3_Release(done);
	AS3_Release(exec);
	AS3_Release(call);
	AS3_Release(fork);
	AS3_Release(forkStorage);
	AS3_Release(trigger);
	AS3_LibInit(result);
	return 0;
}
